/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"
#include "PartIdList.h"
#include "IntBtree.h"
#include <E/sm_btree.h>
extern persistent btree<int> BuildDateIdx;

//////////////////////////////////////////////////////////////////
//
// Reorganization Test #2 - Randomly delete and reinsert as many
// atomic parts per composite part as each one has, interleaving
// reorganizations of all of the composite parts, to test out the
// deletion/reallocation handling and performance characteristics
// of the system
//
//////////////////////////////////////////////////////////////////


extern char*	types[NumTypes];
extern int	debugMode;


//////////////////////////////////////////////////////////////////
//
// Routine to do Reorg2
//
//////////////////////////////////////////////////////////////////


int reorg2()
{
    AssocItr* cpItr; // iterator for traversing AllCompParts collection
    CompositePart* cp;

    // for as many atomic parts as each composite part contains,
    // iterate through all the composite parts, deleting and then
    // recreating one of the atomic parts from its atomic parts graph
    // (and count the number of atomic parts processed in this way)

    int count = 0;
    for (int i = 0; i < NumAtomicPerComp; i++) 
    {
    	cpItr = new AssocItr(&AllCompParts);
    	cp = (CompositePart*) cpItr->next();
    	while (cp != NULL)
    	{
            if (debugMode) {
                printf("    In Reorg2, partId = %d:\n", cp->id);
            }
            count += cp->reorg2();
    	    cp = (CompositePart*) cpItr->next();
        }
	delete cpItr;
    }
    return count;
}



//////////////////////////////////////////////////////////////////
//
// Composite Part Methods for Reorg2
//
//////////////////////////////////////////////////////////////////


int CompositePart::reorg2()
{
    AtomicPart* victim;
    AssocItr* connI;
    PartIdList toList;
    PartIdList fromList;
    Connection* cn;

    if (debugMode) {
        printf("CompositePart::Reorg2(id = %d)\n", id);
    }

    // first select and locate a victim atomic part

    int lowId = rootPart->id;
    int victimId = lowId + (int) (random() % NumAtomicPerComp);
    victim = (AtomicPart*) AtomicPartIdx->lookup(victimId);

    // next record the affected to and from connections

    // first do the to connections
    connI = new AssocItr(victim->to);
    cn = (Connection*) connI->next();  // get first connection record
    while (cn != NULL)
    {
	toList.insert(cn->to->id);
    	cn = (Connection*) connI->next();  // get first connection record
    }
    delete connI;

    // then do the from connections
    connI = new AssocItr(victim->from);
    cn = (Connection*) connI->next();  // get first connection record
    while (cn != NULL)
    {
	toList.insert(cn->from->id);
    	cn = (Connection*) connI->next();  // get first connection record
    }
    delete connI;

    // now do the actual part deletion

    if (debugMode) {
        printf("\tReorg2: Deleting atomic part %d\n", victim->id);
    }
    AtomicPartIdx->remove(victim->id); // remove part from index 
    delete victim;

    // time to (approximately) recreate the deleted part
    victim = new  AtomicPart (victimId, this);

    // stick the id of the part into the hash index
    AtomicPartIdx->insert(victimId, victim);
    // add the buildDate to the BuildDateIdx
    BuildDateIdx.insert(victim->buildDate, victim);

    if (victimId == lowId) rootPart = victim;
    
    if (debugMode) {
        printf("\tReorg2: Recreated atomic part %d\n", victim->id);
    }

    // finally, connect the part back up to the other parts again

    int partId;
    AtomicPart* part;

    while (!toList.empty()) {
	partId = toList.remove();
	part = (AtomicPart*) AtomicPartIdx->lookup(partId);
	cn = new Connection (victim, part);
    }

    while (!fromList.empty()) {
	partId = fromList.remove();
	part = (AtomicPart*) AtomicPartIdx->lookup(partId);
	cn = new Connection (part, victim);
    }

    return 1;
}
